/*! 
    \file vector.h
    \brief A basic 3D math framework
    
    Contains classes to handle vectors, lines, rotations and planes
*/ 

#ifndef _VECTOR_H
#define _VECTOR_H

#include <math.h>
#include "compiler.h"
//! PI the circle-constant
#define PI 3.14159265359f

//! 3D Vector
/**
	Class to handle 3D Vectors
*/
class Vector {

  public:
 
  float x; //!< The x Coordinate of the Vector.
  float y; //!< The y Coordinate of the Vector.
  float z; //!< The z Coordinate of the Vector.

  Vector (float x, float y, float z) : x(x), y(y), z(z) {}  //!< assignment constructor
  Vector () : x(0), y(0), z(0) {}
  ~Vector () {}

  inline Vector operator+ (const Vector& v) const { return Vector(x + v.x, y + v.y, z + v.z); }
  inline const Vector& operator+= (const Vector& v) {this->x += v.x; this->y += v.y; this->z += v.z; return *this;}
  inline Vector operator- (const Vector& v) const { return Vector(x - v.x, y - v.y, z - v.z); }
  inline const Vector& operator-= (const Vector& v) {this->x -= v.x; this->y -= v.y; this->z -= v.z; return *this;}
  inline float operator* (const Vector& v) const { return x * v.x + y * v.y + z * v.z; }
  inline const Vector& operator*= (const Vector& v) {this->x *= v.x; this->y *= v.y; this->z *= v.z; return *this;}
  inline Vector operator* (float f) const { return Vector(x * f, y * f, z * f); }
  inline const Vector& operator*= (float f) {this->x *= f; this->y *= f; this->z *= f; return *this;}
  Vector operator/ (float f) const;
  inline const Vector& operator/= (float f) {this->x /= f; this->y /= f; this->z /= f; return *this;}
  inline const Vector& operator= (const Vector& v) {this->x = v.x; this->y = v.y; this->z = v.z; return *this;}
  float dot (const Vector& v) const;
  inline Vector cross (const Vector& v) const { return Vector(y * v.z - z * v.y, z * v.x - x * v.z, x * v.y - y * v.x ); }
  void scale(const Vector& v);
  inline float len() const { return sqrt (x*x+y*y+z*z); }
  inline void normalize() { 
                      float l = len(); 
		      if( unlikely(l == 0.0)) 
			{ 
			  // Prevent divide by zero
			  return;
			}
		      x = x / l;
		      y = y / l;
		      z = z / l; 
                    }
  Vector getNormalized();
  Vector abs();

  void debug() const;
};

float angleDeg (const Vector& v1, const Vector& v2);
float angleRad (const Vector& v1, const Vector& v2);

//! Quaternion
/**
	Class to handle 3-dimensional rotation efficiently
*/
class Quaternion
{
 public:
  Vector v;	//!< Imaginary Vector
  float w;        //!< Real part of the number
  
  inline Quaternion () { w = 1; v = Vector(0,0,0); }
  inline Quaternion (const Vector& v, float f) { this->w = f; this->v = v; }
  Quaternion (float m[4][4]);
  inline Quaternion (float angle, const Vector& axis) { w = cos(angle/2); v = axis * sin(angle/2); }
  Quaternion (const Vector& dir, const Vector& up);
  Quaternion (float roll, float pitch, float yaw);
  Quaternion operator/ (const float& f) const;
  inline const Quaternion operator/= (const float& f) {*this = *this / f; return *this;}
  Quaternion operator* (const float& f) const;
  inline const Quaternion operator*= (const float& f) {*this = *this * f; return *this;}
  Quaternion operator* (const Quaternion& q) const;
  inline const Quaternion operator*= (const Quaternion& q) {*this = *this * q; return *this;}
  inline Quaternion operator+ (const Quaternion& q) const { return Quaternion(q.v + v, q.w + w); }
  inline const Quaternion& operator+= (const Quaternion& q) {this->v += q.v; this->w += q.w; return *this;}
  inline Quaternion operator- (const Quaternion& q) const { return Quaternion(q.v - v, q.w - w); }
  inline const Quaternion& operator-= (const Quaternion& q) {this->v -= q.v; this->w -= q.w; return *this;}
  inline Quaternion operator= (const Quaternion& q) {this->v = q.v; this->w = q.w; return *this;}
  Quaternion conjugate () const {  Quaternion r(*this);
  r.v = Vector() - r.v;
  return r;}
  Quaternion inverse () const;
  Vector apply (Vector& f) const;
  float norm () const;
  void matrix (float m[4][4]) const;
  
  void debug();
};

Quaternion quatSlerp(const Quaternion& from, const Quaternion& to, float t);



//! 3D rotation (OBSOLETE)
/**
  Class to handle 3-dimensional rotations.
  Can create a rotation from several inputs, currently stores rotation using a 3x3 Matrix
*/
class Rotation {
  public:
  
  float m[9]; //!< 3x3 Rotation Matrix
  
  Rotation ( const Vector& v);
  Rotation ( const Vector& axis, float angle);
  Rotation ( float pitch, float yaw, float roll);
  Rotation ();
  ~Rotation () {}
  
  Rotation operator* (const Rotation& r);
  
  void glmatrix (float* buffer);
};

//!< Apply a rotation to a vector
Vector rotateVector( const Vector& v, const Rotation& r);

//! 3D line
/**
  Class to store Lines in 3-dimensional space

  Supports line-to-line distance measurements and rotation
*/
class Line
{
  public:
  
  Vector r;   //!< Offset
  Vector a;   //!< Direction
  
  Line ( Vector r, Vector a) : r(r), a(a) {}  //!< assignment constructor
  Line () : r(Vector(0,0,0)), a(Vector (1,1,1)) {}
  ~Line () {}
  
  float distance (const Line& l) const;
  float distancePoint (const Vector& v) const;
  Vector* footpoints (const Line& l) const;
  float len () const;
  
  void rotate(const Rotation& rot);
};

//! 3D plane
/**
  Class to handle planes in 3-dimensional space
  
  Critical for polygon-based collision detection
*/
class Plane
{
  public:
  
  Vector n;   //!< Normal vector
  float k;    //!< Offset constant
  
  Plane (Vector a, Vector b, Vector c);
  Plane (Vector norm, Vector p);
  Plane (Vector n, float k) : n(n), k(k) {} //!< assignment constructor
  Plane () : n(Vector(1,1,1)), k(0) {}
  ~Plane () {}
  
  Vector intersectLine (const Line& l) const;
  float distancePoint (const Vector& p) const;
  float locatePoint (const Vector& p) const;
};

#endif /* _VECTOR_H */
